/* This is a stripped-down version of Paul Garrett's complex class */
/* http://www.math.umn.edu/~garrett/a02/Complex.java               */

import java.awt.*;

public class Complex {

    /*
     * Paul Garrett, 18 July 98
     *
     * basic Complex number arithmetic, linear fractional transformations, etc.
     */

    double x, y;

    final static Complex I = new Complex(0D, 1D); // convenient constant
    final static Complex U = new Complex(1D, 0D); // <-- added by DD

    public Complex(double x, double y) {
	this.x = x;
	this.y = y;
    }

    public Complex(double x) {
	this(x,0D);
    }

    /**********************************/

    public final static double re(Complex alpha) {
	return alpha.x;
    }

    public final double re() {
	return Complex.re(this);
    }

    /********************/

    public static final double im(Complex alpha) {
	return alpha.y;
    }

    public final double im() {
	return Complex.im(this);
    }

    /********************/
  
    public static final Complex add(Complex alpha, Complex beta) {
	return new Complex(alpha.re() + beta.re(), alpha.im() + beta.im());
    }

    public final Complex add(Complex beta) {
	return Complex.add(this, beta);
    }

    /*********************/

    public static final Complex sub(Complex alpha, Complex beta) {
	return new Complex(alpha.re() - beta.re(), alpha.im() - beta.im());
    }

    public final Complex sub(Complex beta) {
	return Complex.sub(this, beta);
    }

    /*********************/
    
    public static final Complex mul(Complex alpha, Complex beta) {
	return new Complex(alpha.re() * beta.re() - alpha.im() * beta.im(),
			   alpha.re() * beta.im() + alpha.im() * beta.re());
    }

    public final Complex mul(Complex beta) {
	return Complex.mul(this, beta);
    }

    /**********************/
  
    public static final Complex realMul(double r, Complex alpha) {
	return new Complex(r * alpha.x, r * alpha.y);
    }

    public final Complex realMul(double r) {
	return Complex.realMul(r, this);
    }
  
    /************************/

    public static final Complex realDiv(double r, Complex alpha) {
	return new Complex(alpha.re()/r, alpha.im()/r);
    }
  
    public final Complex realDiv(double r) {
	return Complex.realDiv(r, this);
    }

    /****************************/
  
    public static final Complex conj(Complex alpha) {
	return new Complex(alpha.re(), -alpha.im());
    }
  
    public final Complex conj() {
	return Complex.conj(this);
    }

    /**************************/

    public static final double normSq(Complex alpha) {
	return alpha.x * alpha.x + alpha.y * alpha.y;
    }

    public final double normSq() {
	return Complex.normSq(this);
    }

    /*********************/

    public static final double norm(Complex alpha) {
	return Math.sqrt(alpha.x*alpha.x + alpha.y*alpha.y);
    }

    public final double norm() {
	return Complex.norm(this);
    }

    /********************/

    public static final Complex inv(Complex alpha) {
	double x = alpha.x;
	double y = alpha.y;
	double nsq = x*x+y*y;
	if (nsq <= .0000001) { //__EDIT__??
	    nsq = .0000001;
	}
	return new Complex(x/nsq, -y/nsq);
    }

    public final Complex inv() {
	return Complex.inv(this);
    }

    /**********************/

    public static final Complex div(Complex alpha, Complex beta) {
	return alpha.mul(beta.inv());
    }

    public final Complex div(Complex beta) {
	return Complex.div(this, beta);
    }

    /************************/

    public static final Complex neg(Complex alpha) {
	return new Complex(-alpha.re(), -alpha.im());
    }

    public final Complex neg() {
	return Complex.neg(this);
    }

    /*********************/

    /*
     * This returns and "arg" between -PI and +PI
     *
     * NB arcTangent is "stabilized" against division by 0
     */
    public static final double arg(Complex alpha) {
	return arcTangent2(alpha.y, alpha.x);
    }

    public final double arg() {
	return arcTangent2(this.y, this.x); // see below: custom atan2()
    }

    /*****************************
     *
     * Apply a linear fractional transformation to complex number
     */

    public static final Complex LFT(Complex[][] g, Complex z) {
	return g[0][0].mul(z).add(g[0][1]).div(g[1][0].mul(z).add(g[1][1]));
	// just az+b / cz+d form...
    }

    public final Complex LFT(Complex[][] g) {
	return Complex.LFT(g, this);
    }

    /**********************************************
     * 
     * For command-line output, such as diagnostics
     */ 

    public static final void print(Complex z) {
	System.out.print("(" + z.re() + ") + i(" + z.im() + ")");
    }

    public final void print() {
	System.out.print("(" + this.re() + ") + i(" + this.im() + ")");
    }

    public static final void println(Complex z) {
	System.out.println("(" + z.re() + ") + i(" + z.im() + ")");
    }

    public final void println() {
	System.out.println("(" + this.re() + ") + i(" + this.im() + ")");
    }

    /****************************	 
     *
     * Some compilers, such as VisualAge1.1.2, do not have a good atan2()
     */
  
    public static final double arcTangent2(double y, double x) {
	if (Math.abs(x) < .0001 * Math.abs(y)) { // effectively 0!?
	    if (y > .0001) {
		return Math.PI/2;
	    }
	    else {
		return -Math.PI/2;
	    }
	}
	else { // case that x is not too tiny by comparison to y
	    double angle = Math.atan(Math.abs(y)/Math.abs(x)); // not Math.atan2()
	    if (y<0 && x>=0) {
		angle = -angle;
	    }
	    else if (y>=0 && x<0) {
		angle = Math.PI - angle;
	    }
	    else if (y<=0 && x<0) {
		angle = angle - Math.PI;
	    }
	    return angle;
	}
    }

    public final static int radToDeg(double rad) {
	return ((((int) Math.round(180 * rad / Math.PI)) % 360) + 360) % 360;
    }

    public final static Complex pointToComplex(Point p, double scale) {
	double x,y;
	x = (((double) p.x) / scale) - 1;
	y = (((double) p.y) / scale) - 1;
	return new Complex(x,y); // in unit disk
    }

    public final static Point complexToPoint(Complex z, double scale) {
	int x,y;
	x = (int) Math.round(scale + scale * z.re());
	y = (int) Math.round(scale + scale * z.im());
	return new Point(x,y);
    }

    // non-static version of previous:
    public final Point complexToPoint(double scale) {
	return complexToPoint(this, scale);
    }


    /*
     * Standard complex exponentiation, in polar coordinates
     */

    public static final Complex cxExp(double r, double theta) {
	return new Complex(r*Math.cos(theta), r*Math.sin(theta));
    }


    // My additions - David Dumas
    public static final Complex sqrt(Complex z) {
	return Complex.cxExp(z.norm(), 0.5 * z.arg());
    }

    public Complex sqrt() {
	return sqrt(this);
    }

  
    public static final Complex realAdd(Complex alpha, double x) {
	return new Complex(alpha.re() + x, alpha.im());
    }

    public final Complex realAdd(double x) {
	return Complex.realAdd(this, x);
    }
} // End of Complex class
