import java.awt.*;
import java.awt.geom.*;

class PoincareSegment {
    static final double TINY = 0.000001;
    Complex z, w, c;
    double r, start, sweep;
    boolean isDiam;

    // CONSTRUCTORS

    public PoincareSegment(double x1, double y1, double x2, double y2) {
	z = new Complex(x1,y1);
	w = new Complex(x2,y2);
	compute();
    }

    public PoincareSegment(double zcoord[], double wcoord[]) {
	z = new Complex(zcoord[0],zcoord[1]);
	w = new Complex(wcoord[1],wcoord[1]);
	compute();
    }

    public PoincareSegment(Complex z2, Complex w2) {
	z = z2;
	w = w2;
	compute();
    }

    // Compute the Euclidean circular arc representing this geodesic

    private void compute()
    {
	Complex d = z.mul(w.conj());
	if (Math.abs(d.im()) < TINY*(1.0 + Complex.norm(d))) {
	    // System.out.println("Nearly a diameter.");
	    isDiam = true;
	} else {
	    isDiam = false;
	}
	c = getArcCenter(z,w);
	r = Complex.norm(z.sub(c));
	double a[] = getArcAngles(c,z,w);
	start = a[0];
	sweep = a[1];
    }

    // GETTERS

    public Complex getStart()
    {
	return z;
    }

    public Complex getEnd()
    {
	return w;
    }

    public boolean isDiameter()
    {
	return isDiam;
    }

    public double getRadius()
    {
	return r;
    }

    public Complex getCenter()
    {
	return c;
    }

    public double getStartAngle()
    {
	return start;
    }

    public double getEndAngle()
    {
	return (start+sweep);
    }

    public double getSweep()
    {
	return sweep;
    }

    public PoincareSegment getForwardRay()
    {
	return new PoincareSegment(z,getIdealEnd());
    }

    public PoincareSegment getBackwardRay()
    {
	return new PoincareSegment(getIdealStart(),z);
    }

    public PoincareSegment getGeodesic()
    {
	return new PoincareSegment(getIdealStart(),getIdealEnd());
    }

    public PoincareSegment getReverse()
    {
	return new PoincareSegment(w,z);
    }

    //----------------------------------------------------------------------
    // Adapted from Paul Garrett's complex class       
    // http://www.math.umn.edu/~garrett/a02/Complex.java 

    // Move an angle into range [0,2Pi]
    public static double reduceAngle(double a) {
	while (a > 2 * Math.PI) {
	    a -= 2 * Math.PI;
	}
	while (a < 0) {
	    a += 2*Math.PI;
	}
	return a;
    }

    public static Complex getArcCenter(Complex z, Complex w) {
	Complex num = z.realMul(1+w.normSq()).sub(w.realMul(1+z.normSq()));
	Complex denom = z.mul(w.conj()).sub(z.conj().mul(w));
	Complex c = num.div(denom);
	return c;
    }

    public static double[] getArcAngles(Complex c, Complex z, Complex w) {
	double za = z.sub(c).arg();
	double wa = w.sub(c).arg();
	double startAngle, arcAngle;
	if (za >= wa) {
	    if (za - wa <= Math.PI) {
		startAngle = reduceAngle(wa);
		arcAngle = reduceAngle(za - wa);
	    }
	    else { // za - wa > Math.PI
		startAngle = reduceAngle(za);
		arcAngle = reduceAngle(wa - za);
	    }
	}
	else { // wa > za
	    if (wa - za > Math.PI) {
		startAngle = reduceAngle(wa);
		arcAngle = reduceAngle(za - wa);
	    }
	    else {
		startAngle = reduceAngle(za);
		arcAngle = reduceAngle(wa - za);
	    }
	}
	double[] angles = {startAngle, arcAngle};
	return angles;
    }

    // End methods adapted from Garrett
    //----------------------------------------------------------------------

    // ENDPOINTS
    public Complex getIdealEnd()
    {
	return getEndpoint(z,w);
    }

    public Complex getIdealStart()
    {
	return getEndpoint(w,z);
    }

    // SIDES

    public int getSide(Complex p)
    {
	return Poincare.side(z,w,p);
    }

    public double getLength()
    {
	return Poincare.dist(z,w);
    }

    // Return the ideal endpoint of the ray from z to w
    public static Complex getEndpoint(Complex z, Complex w)
    {
	// First convert z and w to the Klein model
	Complex zk = Poincare.toKlein(z);
	Complex wk = Poincare.toKlein(w);
	// d is the displacement vector of wk from xk
	Complex d = wk.sub(zk);
	// Now calculate t>0 for which (wk + t d) has norm 1
	double e = zk.mul(d.conj()).re();
	double disc = Math.sqrt(e * e + d.normSq() * (1.0 - zk.normSq()));
	double t = (disc - e) / d.normSq();
	// And return (wk + t d)
	return (zk.add(d.realMul(t)));
    }

    // Does segment S intersect the geodesic containing this one?
    public boolean intersects(PoincareSegment s)
    {
	return ((getSide(s.getStart()) * getSide(s.getEnd())) == -1);
    }

    // This is a hack; find the intersection point of s with the 
    // COMPLETE GEODESIC containing this.
    public Complex intersection(PoincareSegment s)
    {
	Complex zk = Poincare.toKlein(z);
	Complex wk = Poincare.toKlein(w);
	Complex pk = Poincare.toKlein(s.getStart());
	Complex qk = Poincare.toKlein(s.getEnd());
	return Poincare.fromKlein(Geom.intersection(zk,wk,pk,qk));
    }

    public Complex reflect(Complex p)
    {
	Complex disp = p.sub(c);
	double factor = (r * r) / disp.normSq();
	return c.add(disp.realMul(factor));
    }

    public PoincareSegment reflect(PoincareSegment s)
    {
	return new PoincareSegment(reflect(s.getStart()),reflect(s.getEnd()));
    }

    // GRAPHICS
    public void drawOn(Graphics2D g)
    {
	g.draw(getShape());
    }

    public Shape getShape()
    {
	if (!isDiam) {
	    return new Arc2D.Double(c.re()-r,c.im()-r,2.0*r,2.0*r,-180.0*start/Math.PI,-180.0*sweep/Math.PI,Arc2D.OPEN);
	} else {
	    return new Line2D.Double(z.re(),z.im(),w.re(),w.im());
	}
    }

    // For test/debug
    public void println() {
	System.out.print("z = ");
	z.println();
	System.out.print("w = ");
	w.println();
    }

}


//======================================================================
//                             OLD JUNK
//
//     public GeneralPath getPathOld()
//     {
// 	return getPathOld(25);
//     }

//     public GeneralPath getPathOld(int nseg)
//     {
// 	return getArc(c,r,start,sweep,nseg);
//     }

//     public static final GeneralPath getArcStub(Complex center, double r, double start, double angle, int nseg)
//     {
// 	Complex p = center.add(Complex.cxExp(r, start));
// 	GeneralPath path = new GeneralPath();
// 	path.moveTo((float)p.re(),(float)p.im());
// 	p = center.add(Complex.cxExp(r, start + angle));
// 	path.lineTo((float)p.re(),(float)p.im());
// 	return path;
//     }
	

//     public static final GeneralPath getArc(Complex center, double r, double start, double angle, int nseg)
//     {
// 	double inc = angle/nseg;
// 	Complex p = center.add(Complex.cxExp(r, start));
// 	GeneralPath path = new GeneralPath();
// 	path.moveTo((float)p.re(),(float)p.im());
// 	for (int i=0; i<=nseg; i++) {
// 	    p = center.add(Complex.cxExp(r, start + i*inc));
// 	    path.lineTo((float)p.re(),(float)p.im());
// 	}
// 	return path;
//     }
